/* $$ Clover has instrumented this file $$ */// Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.tapestry.binding;

import java.util.Map;

import ognl.Ognl;
import ognl.OgnlException;
import ognl.TypeConverter;

import org.apache.hivemind.ClassResolver;
import org.apache.hivemind.Location;
import org.apache.tapestry.BindingException;
import org.apache.tapestry.IComponent;
import org.apache.tapestry.Tapestry;
import org.apache.tapestry.services.ExpressionCache;
import org.apache.tapestry.services.ExpressionEvaluator;
import org.apache.tapestry.spec.BeanLifecycle;
import org.apache.tapestry.spec.IBeanSpecification;
import org.apache.tapestry.spec.IApplicationSpecification;
import org.apache.tapestry.util.StringSplitter;

/**
 * Implements a dynamic binding, based on getting and fetching values using JavaBeans property
 * access. This is built upon the <a href="http://www.ognl.org">OGNL </a> library.
 * <p>
 * <b>Optimization of the Expression </b>
 * <p>
 * There's a lot of room for optimization here because we can count on some portions of the
 * expression to be effectively static. Note that we type the root object as {@link IComponent}. We
 * have some expectations that certain properties of the root (and properties reachable from the
 * root) will be constant for the lifetime of the binding. For example, components never change
 * thier page or container. This means that certain property prefixes can be optimized:
 * <ul>
 * <li>page
 * <li>container
 * <li>components. <i>name </i>
 * </ul>
 * <p>
 * This means that once an ExpressionBinding has been triggered, the {@link #toString()}method may
 * return different values for the root component and the expression than was originally set.
 * <p>
 * <b>Identifying Invariants </b>
 * <p>
 * Most expressions are fully dynamic; they must be resolved each time they are accessed. This can
 * be somewhat inefficient. Tapestry can identify certain paths as invariant:
 * <ul>
 * <li>A component within the page hierarchy
 * <li>An {@link org.apache.tapestry.IAsset}from then assets map (property <code>assets</code>)
 * <li>A {@link org.apache.tapestry.IActionListener}from the listener map (property
 * <code>listeners</code>)
 * <li>A bean with a {@link org.apache.tapestry.spec.BeanLifecycle#PAGE}lifecycle (property
 * <code>beans</code>)
 * <li>A binding (property <code>bindings</code>)
 * </ul>
 * <p>
 * These optimizations have some inherent dangers; they assume that the components have not
 * overidden the specified properties; the last one (concerning helper beans) assumes that the
 * component does inherit from {@link org.apache.tapestry.AbstractComponent}. If this becomes a
 * problem in the future, it may be necessary to have the component itself involved in these
 * determinations.
 * 
 * @author Howard Lewis Ship
 * @since 2.2
 */

public class ExpressionBinding extends AbstractBinding
{public static com.cortexeb.tools.clover.d __CLOVER_52_0 = com.cortexeb.tools.clover.aq.getRecorder(new char[] {67,58,92,119,111,114,107,115,112,97,99,101,92,106,97,107,97,114,116,97,45,116,97,112,101,115,116,114,121,92,102,114,97,109,101,119,111,114,107,92,116,97,114,103,101,116,92,99,108,111,118,101,114,45,100,98},1097439627296L);
    /**
     * The root object against which the nested property name is evaluated.
     */

    private IComponent _root;

    /**
     * The OGNL expression, as a string.
     */

    private String _expression;

    /**
     * If true, then the binding is invariant, and cachedValue is the ultimate value.
     */

    private boolean _invariant = false;

    /**
     * Stores the cached value for the binding, if invariant is true.
     */

    private Object _cachedValue;

    /**
     * Parsed OGNL expression.
     */

    private Object _parsedExpression;

    /**
     * Flag set once the binding has initialized. _cachedValue, _invariant and _final value for
     * _expression are not valid until after initialization.
     */

    private boolean _initialized;

    /**
     * @since 3.1
     */

    private ExpressionEvaluator _evaluator;

    /** @since 3.1 */

    private ExpressionCache _cache;

    /**
     * Creates a {@link ExpressionBinding}from the root object and an OGNL expression.
     */

    public ExpressionBinding(IComponent root, String expression, Location location,
            ExpressionEvaluator evaluator, ExpressionCache cache)
    {
        super(location);__CLOVER_52_0.S[12139]++;try { __CLOVER_52_0.M[2954]++;

        __CLOVER_52_0.S[12140]++;_root = root;
        __CLOVER_52_0.S[12141]++;_expression = expression;
        __CLOVER_52_0.S[12142]++;_evaluator = evaluator;
        __CLOVER_52_0.S[12143]++;_cache = cache;
    } finally { }}

    public String getExpression()
    {try { __CLOVER_52_0.M[2955]++;
        __CLOVER_52_0.S[12144]++;return _expression;
    } finally { }}

    public IComponent getRoot()
    {try { __CLOVER_52_0.M[2956]++;
        __CLOVER_52_0.S[12145]++;return _root;
    } finally { }}

    /**
     * Gets the value of the property path, with the assistance of a OGNL.
     * 
     * @throws BindingException
     *             if an exception is thrown accessing the property.
     */

    public Object getObject()
    {try { __CLOVER_52_0.M[2957]++;
        __CLOVER_52_0.S[12146]++;initialize();

        __CLOVER_52_0.S[12147]++;if ((((_invariant) && (++__CLOVER_52_0.CT[2087] != 0)) || (++__CLOVER_52_0.CF[2087] == 0))){
            __CLOVER_52_0.S[12148]++;return _cachedValue;}

        __CLOVER_52_0.S[12149]++;return resolveProperty();
    } finally { }}

    private Object resolveProperty()
    {try { __CLOVER_52_0.M[2958]++;
        __CLOVER_52_0.S[12150]++;try
        {
            __CLOVER_52_0.S[12151]++;return _evaluator.readCompiled(_root, _parsedExpression);
        }
        catch (Throwable t)
        {
            __CLOVER_52_0.S[12152]++;throw new BindingException(t.getMessage(), this, t);
        }
    } finally { }}

    /**
     * Returns true if the binding is expected to always return the same value.
     */

    public boolean isInvariant()
    {try { __CLOVER_52_0.M[2959]++;
        __CLOVER_52_0.S[12153]++;initialize();

        __CLOVER_52_0.S[12154]++;return _invariant;
    } finally { }}

    public void setBoolean(boolean value)
    {try { __CLOVER_52_0.M[2960]++;
        __CLOVER_52_0.S[12155]++;setObject((((value ) && (++__CLOVER_52_0.CT[2088] != 0)) || (++__CLOVER_52_0.CF[2088] == 0))? Boolean.TRUE : Boolean.FALSE);
    } finally { }}

    public void setInt(int value)
    {try { __CLOVER_52_0.M[2961]++;
        __CLOVER_52_0.S[12156]++;setObject(new Integer(value));
    } finally { }}

    public void setDouble(double value)
    {try { __CLOVER_52_0.M[2962]++;
        __CLOVER_52_0.S[12157]++;setObject(new Double(value));
    } finally { }}

    public void setString(String value)
    {try { __CLOVER_52_0.M[2963]++;
        __CLOVER_52_0.S[12158]++;setObject(value);
    } finally { }}

    /**
     * Sets up the helper object, but also optimizes the property path and determines if the binding
     * is invarant.
     */

    private void initialize()
    {try { __CLOVER_52_0.M[2964]++;
        __CLOVER_52_0.S[12159]++;if ((((_initialized) && (++__CLOVER_52_0.CT[2089] != 0)) || (++__CLOVER_52_0.CF[2089] == 0))){
            __CLOVER_52_0.S[12160]++;return;}

        __CLOVER_52_0.S[12161]++;_initialized = true;

        __CLOVER_52_0.S[12162]++;try
        {
            __CLOVER_52_0.S[12163]++;_parsedExpression = _cache.getCompiledExpression(_expression);
        }
        catch (Exception ex)
        {
            __CLOVER_52_0.S[12164]++;throw new BindingException(ex.getMessage(), this, ex);
        }

        __CLOVER_52_0.S[12165]++;checkForConstant();
    } finally { }}

    private void checkForConstant()
    {try { __CLOVER_52_0.M[2965]++;
        __CLOVER_52_0.S[12166]++;try
        {
            __CLOVER_52_0.S[12167]++;if ((((_evaluator.isConstant(_expression)) && (++__CLOVER_52_0.CT[2090] != 0)) || (++__CLOVER_52_0.CF[2090] == 0))){
            {
                __CLOVER_52_0.S[12168]++;_invariant = true;

                __CLOVER_52_0.S[12169]++;_cachedValue = resolveProperty();
            }}
        }
        catch (Exception ex)
        {
            __CLOVER_52_0.S[12170]++;throw new BindingException(ex.getMessage(), this, ex);
        }
    } finally { }}

    /**
     * Updates the property for the binding to the given value.
     * 
     * @throws BindingException
     *             if the property can't be updated (typically due to an security problem, or a
     *             missing mutator method).
     * @throws ReadOnlyBindingException
     *             if the binding is invariant.
     */

    public void setObject(Object value)
    {try { __CLOVER_52_0.M[2966]++;
        __CLOVER_52_0.S[12171]++;initialize();

        __CLOVER_52_0.S[12172]++;if ((((_invariant) && (++__CLOVER_52_0.CT[2091] != 0)) || (++__CLOVER_52_0.CF[2091] == 0))){
            __CLOVER_52_0.S[12173]++;throw createReadOnlyBindingException(this);}

        __CLOVER_52_0.S[12174]++;try
        {
            __CLOVER_52_0.S[12175]++;_evaluator.writeCompiled(_root, _parsedExpression, value);
        }
        catch (Throwable ex)
        {
            __CLOVER_52_0.S[12176]++;throw new BindingException(ex.getMessage(), this, ex);
        }
    } finally { }}

    /**
     * Returns the a String representing the property path. This includes the
     * {@link IComponent#getExtendedId() extended id}of the root component and the property path
     * ... once the binding is used, these may change due to optimization of the property path.
     */

    public String toString()
    {try { __CLOVER_52_0.M[2967]++;
        __CLOVER_52_0.S[12177]++;StringBuffer buffer = new StringBuffer();

        __CLOVER_52_0.S[12178]++;buffer.append("ExpressionBinding[");
        __CLOVER_52_0.S[12179]++;buffer.append(_root.getExtendedId());

        __CLOVER_52_0.S[12180]++;if ((((_expression != null) && (++__CLOVER_52_0.CT[2092] != 0)) || (++__CLOVER_52_0.CF[2092] == 0))){
        {
            __CLOVER_52_0.S[12181]++;buffer.append(' ');
            __CLOVER_52_0.S[12182]++;buffer.append(_expression);
        }}

        __CLOVER_52_0.S[12183]++;if ((((_invariant) && (++__CLOVER_52_0.CT[2093] != 0)) || (++__CLOVER_52_0.CF[2093] == 0))){
        {
            __CLOVER_52_0.S[12184]++;buffer.append(" cachedValue=");
            __CLOVER_52_0.S[12185]++;buffer.append(_cachedValue);
        }}

        __CLOVER_52_0.S[12186]++;buffer.append(']');

        __CLOVER_52_0.S[12187]++;return buffer.toString();
    } finally { }}

}