/* $$ Clover has instrumented this file $$ */// Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.tapestry;

import java.io.BufferedWriter;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.Stack;

import org.apache.tapestry.util.ContentType;
import org.apache.tapestry.util.text.ICharacterTranslator;
import org.apache.tapestry.util.text.ICharacterTranslatorSource;

/**
 * Abstract base class implementing the {@link IMarkupWriter} interface.
 * This class is used to create a Generic Tag Markup Language (GTML) output.   
 * It is more sophisticated than <code>PrintWriter</code> in that it maintains   
 * a concept hierarchy of open GTML tags. It also supplies a number of other
 * of the features that are useful when creating GTML.
 *
 * Elements are started with the {@link #begin(String)} 
 * or {@link #beginEmpty(String)}
 * methods. Once they are started, attributes for the elements may be set with
 * the various <code>attribute()</code> methods. The element is closed off
 * (i.e., the closing '&gt;' character is written) when any other method
 * is invoked (exception: methods which do not produce output, such as
 * {@link #flush()}). The <code>end()</code> methods end an element,
 * writing an GTML close tag to the output.
 *
 * <p>TBD:
 * <ul>
 * <li>Support XML and XHTML
 *  <li>What to do with Unicode characters with a value greater than 255?
 * </ul>
 *
 * <p>This class is derived from the original class 
 * <code>com.primix.servlet.HTMLWriter</code>,
 * part of the <b>ServletUtils</b> framework available from
 * <a href="http://www.gjt.org/servlets/JCVSlet/list/gjt/com/primix/servlet">The Giant 
 * Java Tree</a>.
 *
 * @author Howard Ship, David Solis
 * @since 0.2.9
 *
 **/

public abstract class AbstractMarkupWriter implements IMarkupWriter
{public static com.cortexeb.tools.clover.d __CLOVER_1_0 = com.cortexeb.tools.clover.aq.getRecorder(new char[] {67,58,92,119,111,114,107,115,112,97,99,101,92,106,97,107,97,114,116,97,45,116,97,112,101,115,116,114,121,92,102,114,97,109,101,119,111,114,107,92,116,97,114,103,101,116,92,99,108,111,118,101,114,45,100,98},1096998272901L);
    /**
     * The encoding to be used should it be omitted in the constructors.
     * This is only used for backward compatibility. New code always provides the encoding.
     */

    private static final String DEFAULT_ENCODING = "utf-8";

    /**
     * The underlying {@link PrintWriter} that output is sent to. 
     *  
     **/

    private PrintWriter _writer;

    /**
     * Indicates whether a tag is open or not. A tag is opened by
     * {@link #begin(String)} or {@link #beginEmpty(String)}.
     * It stays open while calls to the <code>attribute()</code>
     * methods are made. It is closed
     * (the '&gt;' is written) when any other method is invoked.
     *
     **/

    private boolean _openTag = false;

    /**
     *  Indicates that the tag was opened with 
     *  {@link #beginEmpty(String)}, which affects
     *  how the tag is closed (a slash is added to indicate the
     *  lack of a body).  This is compatible with HTML, but reflects
     *  an XML/XHTML leaning.
     * 
     *  @since 2.2
     * 
     **/

    private boolean _emptyTag = false;

    /**
     * A Stack of Strings used to track the active tag elements. Elements are active
     * until the corresponding close tag is written.  The {@link #push(String)} method
     * adds elements to the stack, {@link #pop()} removes them.
     *
     **/

    private Stack _activeElementStack;

    /**
     * The depth of the open tag stack.
     * @see #_activeElementStack
     *
     **/

    private int _depth = 0;

    private char[] _buffer;

    /**
     *  Implemented in concrete subclasses to provide an indication of which
     *  characters are 'safe' to insert directly into the response.  The index
     *  into the array is the character, if the value at the index is false (or the
     *  index out of range), then the character is escaped.
     *
     **/

    private ICharacterTranslator _translator;
    
    private String _contentType;

    /**
     *  Indicates whether {@link #close()} should close the 
     *  underlying {@link PrintWriter}.
     * 
     *  @since 3.0
     * 
     **/

    private boolean _propagateClose = true;

    public String getContentType()
    {try { __CLOVER_1_0.M[61]++;
        __CLOVER_1_0.S[184]++;return _contentType;
    } finally { }}

    abstract public IMarkupWriter getNestedWriter();

    /**
     *  General constructor used by subclasses.
     * 
     *  @param safe an array of flags indicating which characters
     *  can be passed directly through with out filtering.  Characters marked
     *  unsafe, or outside the range defined by safe, are converted to entities.
     *  @param entities a set of prefered entities, unsafe characters with
     *  a defined entity use the entity, other characters are converted
     *  to numeric entities.
     *  @param contentType the MIME type of the content produced by the writer.
     *  @param encoding the encoding of content produced by the writer.
     *  @param stream stream to which content will be written.
     *
     **/

    protected AbstractMarkupWriter(
        ICharacterTranslatorSource translatorSource,
        String contentType,
        String encoding,
        OutputStream stream)
    {try { __CLOVER_1_0.M[62]++;
        __CLOVER_1_0.S[185]++;if ((((translatorSource == null || contentType == null || encoding == null) && (++__CLOVER_1_0.CT[43] != 0)) || (++__CLOVER_1_0.CF[43] == 0))){
            __CLOVER_1_0.S[186]++;throw new IllegalArgumentException(
                Tapestry.getMessage("AbstractMarkupWriter.missing-constructor-parameters"));}

        __CLOVER_1_0.S[187]++;ContentType contentTypeObject = generateFullContentType(contentType, encoding);
        __CLOVER_1_0.S[188]++;_contentType = contentTypeObject.unparse();
        
        __CLOVER_1_0.S[189]++;encoding = contentTypeObject.getParameter("charset");
        __CLOVER_1_0.S[190]++;_translator = translatorSource.getTranslator(encoding);
        
        __CLOVER_1_0.S[191]++;setOutputStream(stream, encoding);
    } finally { }}

    /**
     *  General constructor used by subclasses.
     *  This constructor is left for backward compatibility. It is preferred that 
     *  it is not used since it does not specify an encoding for conversion.
     * 
     *  @param safe an array of flags indicating which characters
     *  can be passed directly through with out filtering.  Characters marked
     *  unsafe, or outside the range defined by safe, are converted to entities.
     *  @param entities a set of prefered entities, unsafe characters with
     *  a defined entity use the entity, other characters are converted
     *  to numeric entities.
     *  @param contentType the type of content produced by the
     *  writer.
     *  @param stream stream to which content will be written.
     **/

    protected AbstractMarkupWriter(
    	ICharacterTranslatorSource translatorSource,
        String contentType,
        OutputStream stream)
    {
        this(translatorSource, contentType);__CLOVER_1_0.S[192]++;try { __CLOVER_1_0.M[63]++;

        __CLOVER_1_0.S[193]++;ContentType contentTypeObject = new ContentType(contentType);
        __CLOVER_1_0.S[194]++;String encoding = contentTypeObject.getParameter("charset");

        __CLOVER_1_0.S[195]++;setOutputStream(stream, encoding);
    } finally { }}

    /**
     *  Creates new markup writer around the underlying {@link PrintWriter}.
     * 
     *  <p>This is primarily used by {@link org.apache.tapestry.engine.TagSupportService},
     *  which is inlcuding content, and therefore this method will not
     *  close the writer when the markup writer is closed.
     * 
     *  @since 3.0
     *  
     **/

    protected AbstractMarkupWriter(
    	ICharacterTranslatorSource translatorSource,
        String contentType,
        PrintWriter writer)
    {
        this(translatorSource, contentType);__CLOVER_1_0.S[196]++;try { __CLOVER_1_0.M[64]++;

        // When the markup writer is closed, the underlying writer
        // is NOT closed.

        __CLOVER_1_0.S[197]++;_propagateClose = false;
        __CLOVER_1_0.S[198]++;_writer = writer;
    } finally { }}

    /**
     *  Special constructor used for nested response writers.
     *  The subclass is responsible for creating the writer.
     * 
     **/

    protected AbstractMarkupWriter(ICharacterTranslatorSource translatorSource, String contentType)
    {try { __CLOVER_1_0.M[65]++;
        __CLOVER_1_0.S[199]++;if ((((translatorSource == null || contentType == null) && (++__CLOVER_1_0.CT[44] != 0)) || (++__CLOVER_1_0.CF[44] == 0))){
            __CLOVER_1_0.S[200]++;throw new IllegalArgumentException(
                Tapestry.getMessage("AbstractMarkupWriter.missing-constructor-parameters"));}

        __CLOVER_1_0.S[201]++;ContentType contentTypeObject = generateFullContentType(contentType, DEFAULT_ENCODING);
        __CLOVER_1_0.S[202]++;_contentType = contentTypeObject.unparse();
        
        __CLOVER_1_0.S[203]++;String encoding = contentTypeObject.getParameter("charset");
        __CLOVER_1_0.S[204]++;_translator = translatorSource.getTranslator(encoding);
    } finally { }}

    /**
     * Ensures that the content type has a charset (encoding) parameter.
     * 
     * @param contentType The content type, e.g. text/html. It may contain a charset parameter.
     * @param encoding The value of the charset parameter of the content type if it is not already present.
     * @return The content type containing a charset parameter, e.g. text/html;charset=utf-8 
     */
    private ContentType generateFullContentType(String contentType, String encoding)
    {try { __CLOVER_1_0.M[66]++;
        __CLOVER_1_0.S[205]++;ContentType contentTypeObject = new ContentType(contentType);
        __CLOVER_1_0.S[206]++;if ((((contentTypeObject.getParameter("charset") == null) && (++__CLOVER_1_0.CT[45] != 0)) || (++__CLOVER_1_0.CF[45] == 0))){
            __CLOVER_1_0.S[207]++;contentTypeObject.setParameter("charset", encoding);}
        __CLOVER_1_0.S[208]++;return contentTypeObject;
    } finally { }}

    protected void setWriter(PrintWriter writer)
    {try { __CLOVER_1_0.M[67]++;
        __CLOVER_1_0.S[209]++;_writer = writer;
    } finally { }}

    protected void setOutputStream(OutputStream stream, String encoding)
    {try { __CLOVER_1_0.M[68]++;
        __CLOVER_1_0.S[210]++;try
        {
            __CLOVER_1_0.S[211]++;OutputStreamWriter owriter;
            __CLOVER_1_0.S[212]++;if ((((encoding != null) && (++__CLOVER_1_0.CT[46] != 0)) || (++__CLOVER_1_0.CF[46] == 0))){
                __CLOVER_1_0.S[213]++;owriter = new OutputStreamWriter(stream, encoding);}
            else{
                __CLOVER_1_0.S[214]++;owriter = new OutputStreamWriter(stream);}
            __CLOVER_1_0.S[215]++;Writer bwriter = new BufferedWriter(owriter);

            __CLOVER_1_0.S[216]++;_writer = new PrintWriter(bwriter);
        }
        catch (UnsupportedEncodingException e)
        {
            __CLOVER_1_0.S[217]++;throw new IllegalArgumentException(
                Tapestry.format("illegal-encoding", encoding));
        }
    } finally { }}

    /**
     * Writes an integer attribute into the currently open tag.
     *
     * <p>TBD: Validate that name is legal.
     *
     * @throws IllegalStateException if there is no open tag.
     *
     **/

    public void attribute(String name, int value)
    {try { __CLOVER_1_0.M[69]++;
        __CLOVER_1_0.S[218]++;checkTagOpen();

        __CLOVER_1_0.S[219]++;_writer.print(' ');
        __CLOVER_1_0.S[220]++;_writer.print(name);
        __CLOVER_1_0.S[221]++;_writer.print("=\"");
        __CLOVER_1_0.S[222]++;_writer.print(value);
        __CLOVER_1_0.S[223]++;_writer.print('"');
    } finally { }}

    /**
     *  Writes a boolean attribute into the currently open tag.
     *
     *  <p>TBD: Validate that name is legal.
     *
     *  @throws IllegalStateException if there is no open tag.
     *
     *  @since 3.0
     *
     **/

    public void attribute(String name, boolean value)
    {try { __CLOVER_1_0.M[70]++;
        __CLOVER_1_0.S[224]++;checkTagOpen();

        __CLOVER_1_0.S[225]++;_writer.print(' ');
        __CLOVER_1_0.S[226]++;_writer.print(name);
        __CLOVER_1_0.S[227]++;_writer.print("=\"");
        __CLOVER_1_0.S[228]++;_writer.print(value);
        __CLOVER_1_0.S[229]++;_writer.print('"');
    } finally { }}

    /**
     *  Writes an attribute into the most recently opened tag. This must be called after
     *  {@link #begin(String)}
     *  and before any other kind of writing (which closes the tag).
     *
     *  <p>The value may be null. A null value will be rendered as an empty string.
     *
     *  <p>Troublesome characters in the value are converted to thier GTML entities, much
     *  like a <code>print()</code> method, with the following exceptions:
     *  <ul>
     *  <li>The double quote (&quot;) is converted to &amp;quot;
     *  <li>The ampersand (&amp;) is passed through unchanged
     *  </ul>
     *
     *  @throws IllegalStateException if there is no open tag.
     *  @param name The name of the attribute to write (no validation
     *  is done on the name).
     *  @param value The value to write.  If null, the attribute
     *  name is written as the value.  Otherwise, the
     *  value is written, 
     **/

    public void attribute(String name, String value)
    {try { __CLOVER_1_0.M[71]++;
        __CLOVER_1_0.S[230]++;checkTagOpen();

        __CLOVER_1_0.S[231]++;_writer.print(' ');

        // Could use a check here that name contains only valid characters

        __CLOVER_1_0.S[232]++;_writer.print(name);
        __CLOVER_1_0.S[233]++;_writer.print("=\"");

        __CLOVER_1_0.S[234]++;if ((((value != null) && (++__CLOVER_1_0.CT[47] != 0)) || (++__CLOVER_1_0.CF[47] == 0))){
        {
            __CLOVER_1_0.S[235]++;int length = value.length();

            __CLOVER_1_0.S[236]++;if ((((_buffer == null || _buffer.length < length) && (++__CLOVER_1_0.CT[48] != 0)) || (++__CLOVER_1_0.CF[48] == 0))){
                __CLOVER_1_0.S[237]++;_buffer = new char[length];}

            __CLOVER_1_0.S[238]++;value.getChars(0, length, _buffer, 0);

            __CLOVER_1_0.S[239]++;safePrint(_buffer, 0, length, true);
        }}

        __CLOVER_1_0.S[240]++;_writer.print('"');

    } finally { }}

    /**
      *  Similar to {@link #attribute(String, String)} but no escaping of invalid elements
      *  is done for the value.
      * 
      *  @throws IllegalStateException if there is no open tag.
      *
      *  @since 3.0
      *
      **/
    public void attributeRaw(String name, String value)
    {try { __CLOVER_1_0.M[72]++;
        __CLOVER_1_0.S[241]++;if ((((value == null) && (++__CLOVER_1_0.CT[49] != 0)) || (++__CLOVER_1_0.CF[49] == 0))){
        {
            __CLOVER_1_0.S[242]++;attribute(name, value);
            __CLOVER_1_0.S[243]++;return;
        }}

        __CLOVER_1_0.S[244]++;checkTagOpen();

        __CLOVER_1_0.S[245]++;_writer.print(' ');

        __CLOVER_1_0.S[246]++;_writer.print(name);

        __CLOVER_1_0.S[247]++;_writer.print("=\"");
        __CLOVER_1_0.S[248]++;_writer.print(value);
        __CLOVER_1_0.S[249]++;_writer.print('"');
    } finally { }}

    /**
     * Closes any existing tag then starts a new element. The new element is pushed
     * onto the active element stack.
     **/

    public void begin(String name)
    {try { __CLOVER_1_0.M[73]++;
        __CLOVER_1_0.S[250]++;if ((((_openTag) && (++__CLOVER_1_0.CT[50] != 0)) || (++__CLOVER_1_0.CF[50] == 0))){
            __CLOVER_1_0.S[251]++;closeTag();}

        __CLOVER_1_0.S[252]++;push(name);

        __CLOVER_1_0.S[253]++;_writer.print('<');
        __CLOVER_1_0.S[254]++;_writer.print(name);

        __CLOVER_1_0.S[255]++;_openTag = true;
        __CLOVER_1_0.S[256]++;_emptyTag = false;
    } finally { }}

    /**
     * Starts an element that will not later be matched with an <code>end()</code>
     * call. This is useful for elements such as &lt;hr;&gt; or &lt;br&gt; that
     * do not need closing tags.
     *
     **/

    public void beginEmpty(String name)
    {try { __CLOVER_1_0.M[74]++;
        __CLOVER_1_0.S[257]++;if ((((_openTag) && (++__CLOVER_1_0.CT[51] != 0)) || (++__CLOVER_1_0.CF[51] == 0))){
            __CLOVER_1_0.S[258]++;closeTag();}

        __CLOVER_1_0.S[259]++;_writer.print('<');
        __CLOVER_1_0.S[260]++;_writer.print(name);

        __CLOVER_1_0.S[261]++;_openTag = true;
        __CLOVER_1_0.S[262]++;_emptyTag = true;
    } finally { }}

    /**
     * Invokes <code>checkError()</code> on the
     *  <code>PrintWriter</code> used to format output.
     **/

    public boolean checkError()
    {try { __CLOVER_1_0.M[75]++;
        __CLOVER_1_0.S[263]++;return _writer.checkError();
    } finally { }}

    private void checkTagOpen()
    {try { __CLOVER_1_0.M[76]++;
        __CLOVER_1_0.S[264]++;if ((((!_openTag) && (++__CLOVER_1_0.CT[52] != 0)) || (++__CLOVER_1_0.CF[52] == 0))){
            __CLOVER_1_0.S[265]++;throw new IllegalStateException(
                Tapestry.getMessage("AbstractMarkupWriter.tag-not-open"));}
    } finally { }}

    /**
     * Closes this <code>IMarkupWriter</code>. Any active elements are closed. The
     * {@link PrintWriter} is then  sent {@link PrintWriter#close()}.
     *
     **/

    public void close()
    {try { __CLOVER_1_0.M[77]++;
        __CLOVER_1_0.S[266]++;if ((((_openTag) && (++__CLOVER_1_0.CT[53] != 0)) || (++__CLOVER_1_0.CF[53] == 0))){
            __CLOVER_1_0.S[267]++;closeTag();}

        // Close any active elements.

        __CLOVER_1_0.S[268]++;while ((((_depth > 0) && (++__CLOVER_1_0.CT[54] != 0)) || (++__CLOVER_1_0.CF[54] == 0))){
        {
            __CLOVER_1_0.S[269]++;_writer.print("</");
            __CLOVER_1_0.S[270]++;_writer.print(pop());
            __CLOVER_1_0.S[271]++;_writer.print('>');
        }}

        __CLOVER_1_0.S[272]++;if ((((_propagateClose) && (++__CLOVER_1_0.CT[55] != 0)) || (++__CLOVER_1_0.CF[55] == 0))){
            __CLOVER_1_0.S[273]++;_writer.close();}

        __CLOVER_1_0.S[274]++;_writer = null;
        __CLOVER_1_0.S[275]++;_activeElementStack = null;
        __CLOVER_1_0.S[276]++;_buffer = null;
    } finally { }}

    /**
     *  Closes the most recently opened element by writing the '&gt;' that ends
     *  it. May write a slash before the '&gt;' if the tag
     *  was opened by {@link #beginEmpty(String)}.
     * 
     *  <p>Once this is invoked, the <code>attribute()</code> methods
     *  may not be used until a new element is opened with {@link #begin(String)} or
     *  or {@link #beginEmpty(String)}.
     **/

    public void closeTag()
    {try { __CLOVER_1_0.M[78]++;
        __CLOVER_1_0.S[277]++;if ((((_emptyTag) && (++__CLOVER_1_0.CT[56] != 0)) || (++__CLOVER_1_0.CF[56] == 0))){
            __CLOVER_1_0.S[278]++;_writer.print('/');}

        __CLOVER_1_0.S[279]++;_writer.print('>');

        __CLOVER_1_0.S[280]++;_openTag = false;
        __CLOVER_1_0.S[281]++;_emptyTag = false;
    } finally { }}

    /**
     * Writes an GTML comment. Any open tag is first closed. 
     * The method takes care of
     * providing the <code>&lt;!--</code> and <code>--&gt;</code>, 
     * including a blank line after the close of the comment.
     *
     * <p>Most characters are valid inside an GTML comment, so no check
     * of the contents is made (much like {@link #printRaw(String)}.
     *
     **/

    public void comment(String value)
    {try { __CLOVER_1_0.M[79]++;
        __CLOVER_1_0.S[282]++;if ((((_openTag) && (++__CLOVER_1_0.CT[57] != 0)) || (++__CLOVER_1_0.CF[57] == 0))){
            __CLOVER_1_0.S[283]++;closeTag();}

        __CLOVER_1_0.S[284]++;_writer.print("<!-- ");
        __CLOVER_1_0.S[285]++;_writer.print(value);
        __CLOVER_1_0.S[286]++;_writer.println(" -->");
    } finally { }}

    /**
     * Ends the element most recently started by {@link #begin(String)}. 
     * The name of the tag
     * is popped off of the active element stack and used to form an GTML close tag.
     *
     * <p>TBD: Error checking for the open element stack empty.
     **/

    public void end()
    {try { __CLOVER_1_0.M[80]++;
        __CLOVER_1_0.S[287]++;if ((((_openTag) && (++__CLOVER_1_0.CT[58] != 0)) || (++__CLOVER_1_0.CF[58] == 0))){
            __CLOVER_1_0.S[288]++;closeTag();}

        __CLOVER_1_0.S[289]++;_writer.print("</");
        __CLOVER_1_0.S[290]++;_writer.print(pop());
        __CLOVER_1_0.S[291]++;_writer.print('>');
    } finally { }}

    /**
     * Ends the most recently started element with the given name. This will
     * also end any other intermediate elements. This is very useful for easily
     * ending a table or even an entire page.
     *
     * <p>TBD: Error check if the name matches nothing on the open tag stack.
     **/

    public void end(String name)
    {try { __CLOVER_1_0.M[81]++;
        __CLOVER_1_0.S[292]++;if ((((_openTag) && (++__CLOVER_1_0.CT[59] != 0)) || (++__CLOVER_1_0.CF[59] == 0))){
            __CLOVER_1_0.S[293]++;closeTag();}

        __CLOVER_1_0.S[294]++;while (true){
        {
            __CLOVER_1_0.S[295]++;String tagName = pop();

            __CLOVER_1_0.S[296]++;_writer.print("</");
            __CLOVER_1_0.S[297]++;_writer.print(tagName);
            __CLOVER_1_0.S[298]++;_writer.print('>');

            __CLOVER_1_0.S[299]++;if ((((tagName.equals(name)) && (++__CLOVER_1_0.CT[60] != 0)) || (++__CLOVER_1_0.CF[60] == 0))){
                __CLOVER_1_0.S[300]++;break;}
        }}
    } finally { }}

    /**
     * Forwards <code>flush()</code> to this <code>AbstractMarkupWriter</code>'s 
     * <code>PrintWriter</code>.
     *
     **/

    public void flush()
    {try { __CLOVER_1_0.M[82]++;
        __CLOVER_1_0.S[301]++;_writer.flush();
    } finally { }}

    /**
     *  Removes the top element from the active element stack and returns it.
     *
     **/

    protected final String pop()
    {try { __CLOVER_1_0.M[83]++;
        __CLOVER_1_0.S[302]++;String result = (String) _activeElementStack.pop();
        __CLOVER_1_0.S[303]++;_depth--;

        __CLOVER_1_0.S[304]++;return result;
    } finally { }}

    /**
     *
     * The primary <code>print()</code> method, used by most other methods.
     *
     * <p>Prints the character array, first closing any open tag. Problematic characters
     * ('&lt;', '&gt;' and '&amp;') are converted to their
     * GTML entities.
     *
     * <p>All 'unsafe' characters are properly converted to either a named
     * or numeric GTML entity.  This can be somewhat expensive, so use
     * {@link #printRaw(char[], int, int)} if the data to print is guarenteed
     * to be safe.
     *
     * <p>Does <em>nothing</em> if <code>data</code> is null.
     *
     * <p>Closes any open tag.
     *
     **/

    public void print(char[] data, int offset, int length)
    {try { __CLOVER_1_0.M[84]++;
        __CLOVER_1_0.S[305]++;if ((((data == null) && (++__CLOVER_1_0.CT[61] != 0)) || (++__CLOVER_1_0.CF[61] == 0))){
            __CLOVER_1_0.S[306]++;return;}

        __CLOVER_1_0.S[307]++;if ((((_openTag) && (++__CLOVER_1_0.CT[62] != 0)) || (++__CLOVER_1_0.CF[62] == 0))){
            __CLOVER_1_0.S[308]++;closeTag();}

        __CLOVER_1_0.S[309]++;safePrint(data, offset, length, false);
    } finally { }}

    /**
     * Prints a single character. If the character is not a 'safe' character,
     * such as '&lt;', then it's GTML entity (named or numeric) is printed instead.
     *
     * <p>Closes any open tag.
     *
     **/

    public void print(char value)
    {try { __CLOVER_1_0.M[85]++;
        __CLOVER_1_0.S[310]++;if ((((_openTag) && (++__CLOVER_1_0.CT[63] != 0)) || (++__CLOVER_1_0.CF[63] == 0))){
            __CLOVER_1_0.S[311]++;closeTag();}

        __CLOVER_1_0.S[312]++;String translation = _translator.translate(value);
        
        __CLOVER_1_0.S[313]++;if ((((translation == null) && (++__CLOVER_1_0.CT[64] != 0)) || (++__CLOVER_1_0.CF[64] == 0))){
        	__CLOVER_1_0.S[314]++;_writer.print(value);}
        else{
        	__CLOVER_1_0.S[315]++;_writer.print(translation);}
    } finally { }}

    /**
     * Prints an integer.
     *
     * <p>Closes any open tag.
     *
     **/

    public void print(int value)
    {try { __CLOVER_1_0.M[86]++;
        __CLOVER_1_0.S[316]++;if ((((_openTag) && (++__CLOVER_1_0.CT[65] != 0)) || (++__CLOVER_1_0.CF[65] == 0))){
            __CLOVER_1_0.S[317]++;closeTag();}

        __CLOVER_1_0.S[318]++;_writer.print(value);
    } finally { }}

    /**
     * Invokes {@link #print(char[], int, int)} to print the string.  Use
     * {@link #printRaw(String)} if the character data is known to be safe.
     *
     * <p>Does <em>nothing</em> if <code>value</code> is null.
     *
     * <p>Closes any open tag.
     *
     * @see #print(char[], int, int)
     *
     **/

    public void print(String value)
    {try { __CLOVER_1_0.M[87]++;
        __CLOVER_1_0.S[319]++;if ((((value == null) && (++__CLOVER_1_0.CT[66] != 0)) || (++__CLOVER_1_0.CF[66] == 0))){
            __CLOVER_1_0.S[320]++;return;}

        __CLOVER_1_0.S[321]++;int length = value.length();

        __CLOVER_1_0.S[322]++;if ((((_buffer == null || _buffer.length < length) && (++__CLOVER_1_0.CT[67] != 0)) || (++__CLOVER_1_0.CF[67] == 0))){
            __CLOVER_1_0.S[323]++;_buffer = new char[length];}

        __CLOVER_1_0.S[324]++;value.getChars(0, length, _buffer, 0);

        __CLOVER_1_0.S[325]++;print(_buffer, 0, length);
    } finally { }}

    /**
     * Closes the open tag (if any), then prints a line seperator to the output stream.
     *
     **/

    public void println()
    {try { __CLOVER_1_0.M[88]++;
        __CLOVER_1_0.S[326]++;if ((((_openTag) && (++__CLOVER_1_0.CT[68] != 0)) || (++__CLOVER_1_0.CF[68] == 0))){
            __CLOVER_1_0.S[327]++;closeTag();}

        __CLOVER_1_0.S[328]++;_writer.println();
    } finally { }}

    /**
     * Prints and portion of an output buffer to the stream.
     * No escaping of invalid GTML elements is done, which
     * makes this more effecient than <code>print()</code>. 
     * Does <em>nothing</em> if <code>buffer</code>
     * is null.
     *
     * <p>Closes any open tag.
     *
     **/

    public void printRaw(char[] buffer, int offset, int length)
    {try { __CLOVER_1_0.M[89]++;
        __CLOVER_1_0.S[329]++;if ((((buffer == null) && (++__CLOVER_1_0.CT[69] != 0)) || (++__CLOVER_1_0.CF[69] == 0))){
            __CLOVER_1_0.S[330]++;return;}

        __CLOVER_1_0.S[331]++;if ((((_openTag) && (++__CLOVER_1_0.CT[70] != 0)) || (++__CLOVER_1_0.CF[70] == 0))){
            __CLOVER_1_0.S[332]++;closeTag();}

        __CLOVER_1_0.S[333]++;_writer.write(buffer, offset, length);
    } finally { }}

    /**
     * Prints output to the stream. No escaping of invalid GTML elements is done, which
     * makes this more effecient than <code>print()</code>. Does <em>nothing</em> 
     * if <code>value</code>
     * is null.
     *
     * <p>Closes any open tag.
     *
     **/

    public void printRaw(String value)
    {try { __CLOVER_1_0.M[90]++;
        __CLOVER_1_0.S[334]++;if ((((value == null) && (++__CLOVER_1_0.CT[71] != 0)) || (++__CLOVER_1_0.CF[71] == 0))){
            __CLOVER_1_0.S[335]++;return;}

        __CLOVER_1_0.S[336]++;if ((((_openTag) && (++__CLOVER_1_0.CT[72] != 0)) || (++__CLOVER_1_0.CF[72] == 0))){
            __CLOVER_1_0.S[337]++;closeTag();}

        __CLOVER_1_0.S[338]++;_writer.print(value);
    } finally { }}

    /**
     *  Adds an element to the active element stack.
     *
     **/

    protected final void push(String name)
    {try { __CLOVER_1_0.M[91]++;
        __CLOVER_1_0.S[339]++;if ((((_activeElementStack == null) && (++__CLOVER_1_0.CT[73] != 0)) || (++__CLOVER_1_0.CF[73] == 0))){
            __CLOVER_1_0.S[340]++;_activeElementStack = new Stack();}

        __CLOVER_1_0.S[341]++;_activeElementStack.push(name);

        __CLOVER_1_0.S[342]++;_depth++;
    } finally { }}

    /**
     * Internal support for safe printing.  Ensures that all characters emitted
     * are safe: either valid GTML characters or GTML entities (named or numeric).
     **/

    private void safePrint(char[] data, int offset, int length, boolean isAttribute)
    {try { __CLOVER_1_0.M[92]++;
        __CLOVER_1_0.S[343]++;int safelength = 0;
        __CLOVER_1_0.S[344]++;int start = offset;

        __CLOVER_1_0.S[345]++;for (int i = 0; (((i < length) && (++__CLOVER_1_0.CT[74] != 0)) || (++__CLOVER_1_0.CF[74] == 0)); i++){
        {
            __CLOVER_1_0.S[346]++;char ch = data[offset + i];

            // Ignore safe characters.  Outside an attribute, quotes
            // are ok and are not escaped.

            __CLOVER_1_0.S[347]++;String translation;
            __CLOVER_1_0.S[348]++;if ((((ch == '"' && !isAttribute) && (++__CLOVER_1_0.CT[75] != 0)) || (++__CLOVER_1_0.CF[75] == 0))){
            	__CLOVER_1_0.S[349]++;translation = null;}
            else{
            	__CLOVER_1_0.S[350]++;translation = _translator.translate(ch);}
            
            __CLOVER_1_0.S[351]++;if ((((translation == null) && (++__CLOVER_1_0.CT[76] != 0)) || (++__CLOVER_1_0.CF[76] == 0))){
            {
                __CLOVER_1_0.S[352]++;safelength++;
                __CLOVER_1_0.S[353]++;continue;
            }}

            // Write the safe stuff.

            __CLOVER_1_0.S[354]++;if ((((safelength > 0) && (++__CLOVER_1_0.CT[77] != 0)) || (++__CLOVER_1_0.CF[77] == 0))){
                __CLOVER_1_0.S[355]++;_writer.write(data, start, safelength);}

            __CLOVER_1_0.S[356]++;_writer.print(translation);

            __CLOVER_1_0.S[357]++;start = offset + i + 1;
            __CLOVER_1_0.S[358]++;safelength = 0;
        }}

        __CLOVER_1_0.S[359]++;if ((((safelength > 0) && (++__CLOVER_1_0.CT[78] != 0)) || (++__CLOVER_1_0.CF[78] == 0))){
            __CLOVER_1_0.S[360]++;_writer.write(data, start, safelength);}
    } finally { }}

}