/* $$ Clover has instrumented this file $$ */// Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.tapestry.callback;

import org.apache.hivemind.ApplicationRuntimeException;
import org.apache.tapestry.IExternalPage;
import org.apache.tapestry.IRequestCycle;
import org.apache.tapestry.Tapestry;

/**
 *  A callback for returning to an {@link org.apache.tapestry.IExternalPage}.
 *  <p>
 *  Example usage of <tt>ExternalCallback</tt>: 
 *  <p>
 *  The External page ensure a user is authenticated in the 
 *  {@link org.apache.tapestry.IPage#validate(IRequestCycle)} method. 
 *  If the user is not authenticated, they are redirected to the Login page, after 
 *  setting a callback in the Login page.
 *  <p>
 *  The Login page <tt>formSubmit()</tt> {@link org.apache.tapestry.IActionListener} 
 *  authenticates the user and then invokes {@link ICallback#performCallback(IRequestCycle)} 
 *  to the External page.
 *  <pre>
 *  public class External extends BasePage implements IExternalPage {
 * 
 *      private Integer _itemId;
 *
 *      public void validate(IRequestCycle cycle) throws RequestCycleException {            
 *          Visit visit = (Visit) getVisit();
 *      
 *          if (!visit.isAuthenticated()) {
 *              Login login = (Login) cycle.getPage("Login");
 *
 *              login.setCallback
 *                  (new ExternalCallback(this, cycle.getServiceParameters()));
 *              
 *              throw new PageRedirectException(login);
 *          }            
 *      }
 * 
 *      public void activateExternalPage(Object[] params, IRequestCycle cycle)
 *              throws RequestCycleException {            
 *          _itemId = (Integer) params[0];
 *      }
 *  }
 *
 *  public Login extends BasePage {
 * 
 *      private ICallback _callback;
 *
 *      public void setCallback(ICallback _callback) {
 *          _callback = callback;
 *      }
 *
 *      public void formSubmit(IRequestCycle cycle) {
 *          // Authentication code
 *          ..
 *   
 *          Visit visit = (Visit) getVisit();
 *
 *          visit.setAuthenticated(true);
 *  
 *          if (_callback != null) {
 *              _callback.performCallback(cycle);
 *          }
 *      }
 *  }    
 *  </pre>
 * 
 *  @see org.apache.tapestry.IExternalPage
 *  @see org.apache.tapestry.engine.ExternalService
 *
 *  @author Malcolm Edgar
 *  @since 2.3
 *
 **/

public class ExternalCallback implements ICallback
{public static com.cortexeb.tools.clover.d __CLOVER_57_0 = com.cortexeb.tools.clover.aq.getRecorder(new char[] {67,58,92,119,111,114,107,115,112,97,99,101,92,106,97,107,97,114,116,97,45,116,97,112,101,115,116,114,121,92,102,114,97,109,101,119,111,114,107,92,116,97,114,103,101,116,92,99,108,111,118,101,114,45,100,98},1096998272901L);
    private String _pageName;
    private Object[] _parameters;

    /**
     *  Creates a new ExternalCallback for the named <tt>IExternalPage</tt>.  
     *  The parameters (which may be null) is retained, not copied.
     *
     **/

    public ExternalCallback(String pageName, Object[] parameters)
    {try { __CLOVER_57_0.M[347]++;
        __CLOVER_57_0.S[1420]++;_pageName = pageName;
        __CLOVER_57_0.S[1421]++;_parameters = parameters;
    } finally { }}

    /**
     *  Creates a new ExternalCallback for the page.  The parameters
     *  (which may be null) is retained, not copied.
     *
     **/

    public ExternalCallback(IExternalPage page, Object[] parameters)
    {try { __CLOVER_57_0.M[348]++;
        __CLOVER_57_0.S[1422]++;_pageName = page.getPageName();
        __CLOVER_57_0.S[1423]++;_parameters = parameters;
    } finally { }}

    /**
     *  Invokes {@link IRequestCycle#setPage(String)} to select the previously
     *  identified <tt>IExternalPage</tt> as the response page and activates
     *  the page by invoking <tt>activateExternalPage()</tt> with the callback 
     *  parameters and request cycle.
     *
     **/

    public void performCallback(IRequestCycle cycle)
    {try { __CLOVER_57_0.M[349]++;        
        __CLOVER_57_0.S[1424]++;try
        {
            __CLOVER_57_0.S[1425]++;IExternalPage page = (IExternalPage) cycle.getPage(_pageName);
            
            __CLOVER_57_0.S[1426]++;cycle.activate(page);
    
            __CLOVER_57_0.S[1427]++;page.activateExternalPage(_parameters, cycle);            
        }
        catch (ClassCastException ex)
        {
            __CLOVER_57_0.S[1428]++;throw new ApplicationRuntimeException(
                Tapestry.format("ExternalCallback.page-not-compatible", _pageName),
                ex);
        }
    } finally { }}
    
    public String toString()
    {try { __CLOVER_57_0.M[350]++;
        __CLOVER_57_0.S[1429]++;StringBuffer buffer = new StringBuffer("ExternalCallback[");

        __CLOVER_57_0.S[1430]++;buffer.append(_pageName);
        __CLOVER_57_0.S[1431]++;buffer.append('/');

        __CLOVER_57_0.S[1432]++;if ((((_parameters != null) && (++__CLOVER_57_0.CT[272] != 0)) || (++__CLOVER_57_0.CF[272] == 0))){
        {
            __CLOVER_57_0.S[1433]++;String sep = " ";

            __CLOVER_57_0.S[1434]++;for (int i = 0; (((i < _parameters.length) && (++__CLOVER_57_0.CT[273] != 0)) || (++__CLOVER_57_0.CF[273] == 0)); i++){
            {
                __CLOVER_57_0.S[1435]++;buffer.append(sep);
                __CLOVER_57_0.S[1436]++;buffer.append(_parameters[i]);

                __CLOVER_57_0.S[1437]++;sep = ", ";
            }}
        }}

        __CLOVER_57_0.S[1438]++;buffer.append(']');

        __CLOVER_57_0.S[1439]++;return buffer.toString();
    } finally { }}    
}